from .poly import *

@nb.jit(nopython=True, cache=True)
def get_coeffs(n, l):
    coeffs = np.zeros(l + n - 1)
    for i in range(l + n - 1):
        if l % 2 != i % 2:
            continue
        for j in range(0, i + 1, 2):
            if j > n - 2 or i - j > l:
                continue
            comb_factor = comb((n - 2) // 2, j // 2)
            coeffs[i] += comb_factor * \
                fac2_ratio(l + i - j + n - 4, l - i + j, 
                           n - 4, 2 * i - 2 * j) * 2 ** (i - j)
        coeffs[i] *= (-1) ** ((l - i) // 2)
    return coeffs

@nb.jit(nopython=True, cache=True)
def get_coeffs_recursive(n, l, round_coeffs=False):
    coeffs_j = np.zeros(n - 1)
    coeffs_jp1 = np.zeros(n)
    for k in range(0, n - 1, 2):
        coeffs_j[k] = (-1) ** (k // 2) * comb((n - 2) // 2, k // 2)
        coeffs_jp1[k + 1] = (n - 2) * coeffs_j[k]
    if l == 0:
        return coeffs_j
    if l == 1:
        return coeffs_jp1
    for j in range(l - 1):
        coeffs_jp2 = np.empty(len(coeffs_jp1) + 1)
        coeffs_jp2 = (2 * j + n) / (j + 2) * \
                     np.hstack((np.zeros(1), coeffs_jp1)) - \
                     (j + n - 2) / (j + 2) * \
                     np.hstack((coeffs_j, np.zeros(2)))
        coeffs_j = coeffs_jp1
        coeffs_jp1 = coeffs_jp2
        if round_coeffs:
            np.around(coeffs_jp2, 0, coeffs_jp2)
    return coeffs_jp2 # j == l - 2, so j + 2 == l

@nb.jit(nopython=True, cache=True)
def get_cheby_coeffs(n, l, round_coeffs=False):
    if n % 2:
        coeffs = np.dot(get_coeffs_recursive(n, l, round_coeffs),
                        chebyshev_second_inv_matrix(l + n - 2))
    else:
        coeffs = np.dot(get_coeffs_recursive(n, l, round_coeffs),  
                        chebyshev_first_inv_matrix(l + n - 2))
    return coeffs

@nb.jit(nopython=True, cache=True)
def get_cheby_coeffs_recursive(n, l):
    if l == 0 or l == 1:
        return get_cheby_coeffs(n, l)
    coeffs_j = get_cheby_coeffs(n, 0)
    coeffs_jp1 = get_cheby_coeffs(n, 1)
    for j in range(l - 1):
        coeffs_jp2 = np.empty(len(coeffs_jp1) + 1)
        factor1 = (2 * j + n) / (2 * j + 4)
        factor2 = (j + n - 2) / (j + 2)
        coeffs_jp2[0] = factor1 * coeffs_jp1[1] - factor2 * coeffs_j[0]
        for k in range(1, len(coeffs_jp2) - 2):
             coeffs_jp2[k] = \
                 factor1 * (coeffs_jp1[k + 1] + coeffs_jp1[k - 1]) - \
                 factor2 * coeffs_j[k]
        coeffs_jp2[-1] = factor1 * coeffs_jp1[-1]
        coeffs_jp2[-2] = factor1 * coeffs_jp1[-2]
        coeffs_j = coeffs_jp1
        coeffs_jp1 = coeffs_jp2
    np.round(coeffs_jp2, 14, coeffs_jp2)
    return coeffs_jp2 # j == l - 2, so j + 2 == l

@nb.jit(nopython=True, cache=True)
def fac2_ratio(num, denom1, denom2, denom3):
    assert num == denom1 + denom2 + denom3
    denoms = [denom1, denom2, denom3]
    like_parity_denoms = [denom for denom in denoms
                          if num % 2 == denom % 2]
    max_lp_denom = max(like_parity_denoms)
    for i in range(3):
        if denoms[i] == max_lp_denom:
            max_lp_idx = i
            break
    ratio = 1.
    for k in range(max_lp_denom + 2, num + 2, 2):
        ratio *= k
    for i in range(3):
        if i != max_lp_idx:
            ratio /= factorial2(denoms[i])
    return ratio 

@nb.jit(nopython=True, cache=True)
def factorial2(n):
    fac2 = 1.
    for j in range(n, 0, -2):
        fac2 *= j
    return int(fac2)



from .kernel import *

def sample_spherical_kernel(n, mean, var, hemisphere=False):
    """Sample from the diffusion kernel on the n-dimensional sphere.

    Parameters
    ----------
    n : int
        The dimensionality of the space in which the spherical surface is 
        embedded.
    mean : np.array [N x n]
        The n-dimensional unit vectors to be used as the means of the 
        distributions from which to sample.
    var : np.array [N] 
        Scalar variances of the distributions for each vector.
    hemisphere : bool
        If True, sample on the n-hemisphere instead of the n-sphere.
    """
    assert len(mean) == len(var)
    assert len(mean[0]) == n
    N = len(var)
    normal = np.random.randn(N, n)
    rejections = normal - (normal * mean).sum(axis=1, keepdims=True) * mean
    rejections /= np.linalg.norm(rejections, axis=1, keepdims=True)
    phi = newton_raphson(n, np.random.random(N), var, hemisphere)
    cosines, sines = np.cos(phi).reshape((-1, 1)), np.sin(phi).reshape((-1, 1))
    return mean * cosines + rejections * sines

@nb.jit(nopython=True, cache=True)
def newton_raphson(n, u_phi, var, hemisphere=False):
    """Invert F_angle - u_phi by Newton-Raphson iteration.

    Parameters
    ----------
    n : int
        The dimensionality of the space in which the spherical surface is 
        embedded.
    u_omega : np.array [N]
        CDF values for which to find the corresponding angles phi.
    var : np.array [N] 
        Scalar variances of the distributions for each angle.
    hemisphere : bool
        If True, use PDFs and CDFs from the n-hemisphere instead of the 
        n-sphere.

    Returns
    -------
    phi_opt : np.array [N]
        Array of angles phi for which the CDF is u_phi.
    """
    phi = angle_mean(n, var) 
    phi[var == 0.] = 0.
    counter, d_phi = 0, np.ones_like(phi)
    idxs = np.logical_and(np.abs(d_phi) > 1e-14, var > 0.)
    while np.any(idxs) and counter < 1000:
        pdf, cdf = kernel(n, var[idxs], phi[idxs])
        if hemisphere:
            pdf_rev, cdf_rev = kernel(n, var[idxs], np.pi - phi[idxs])
            pdf = pdf + pdf_rev
            cdf = 1. + cdf - cdf_rev
        d_phi = (cdf - u_phi[idxs]) / pdf
        phi[idxs] = phi[idxs] - d_phi
        idxs[idxs] = np.abs(d_phi) > 1e-14
        counter += 1
    return phi



from .coeffs import *

@nb.jit(nopython=True, cache=True)
def kernel(n, var, phi):
    """Compute the PDF and CDF of the diffusion kernel on the (n-1)D sphere.

       These are functions of phi_{n-1}, the final polar angle in the 
       hyperspherical coordinates of a sphere embedded in nD space. The 
       volume element, sin^{n-2}(phi_{n-1}) is included, such that the 
       PDF multiplied by the volume element integrates to 1 over [0, pi].

    Parameters
    ----------
    n : int
        The dimensionality of the space in which the spherical surface is 
        embedded.
    var : np.array [N]
        Array of variance parameters for which to compute the PDF and CDF 
        of the diffusion kernel at each angle phi.
    phi : np.array [N]
        Array of angles (in radians) at which to compute the PDF and CDF 
        of the diffusion kernel.

    Returns
    -------
    pdf : np.array [N]
        The PDF of the diffusion kernel on the (n-1)D sphere, evaluated at 
        each angle phi.
    cdf : np.array [N]
        The CDF of the diffusion kernel on the (n-1)D sphere, evaluated at 
        each angle phi.
    """
    norm_square = (2. / np.pi) ** (int(n - 3) % 2) * \
                  2 ** (n - 3)
    for j in range(n - 4, 0, -2):
        norm_square *= j ** 2
    for j in range(2 * n - 6, 0, -2):
        norm_square /= j
    pdf, cdf = np.zeros_like(phi), np.zeros_like(phi)
    gegenbauer_l_1 = 1. # lth Gegenbauer polynomial evaluated at 1
    gegenbauer_lp1_1 = n - 2 # l+1th Gegenbauer polynomial evaluated at 1
    coeffs_l = get_cheby_coeffs(n, 0)
    coeffs_lp1 = get_cheby_coeffs(n, 1)
    converged, prev_converged = False, False
    l = 0
    while not (converged and prev_converged):
        prev_converged = converged
        dpdf, dcdf = np.zeros_like(phi), np.zeros_like(phi)
        for k, coeff in enumerate(coeffs_l):
            if not coeff:
                continue
            if n % 2: # n is odd; use a sine series
                dpdf += coeff * np.sin((k + 1) * phi)
                dcdf += coeff * (1 - np.cos((k + 1) * phi)) / (k + 1) 
            elif k > len(coeffs_l) - n: # n is even; use a cosine series
                if k:
                    dpdf += coeff * np.cos(k * phi)
                    dcdf += coeff * np.sin(k * phi) / k
                else:
                    dpdf += coeff
                    dcdf += coeff * phi
        exp_factor = np.exp(-0.5 * l * (l + n - 2) * var)
        dpdf *= 0.5 * (2 * l + n - 2) * \
                gegenbauer_l_1 * norm_square * exp_factor
        dcdf *= 0.5 * (2 * l + n - 2) * \
                gegenbauer_l_1 * norm_square * exp_factor
        pdf += dpdf
        cdf += dcdf
        if np.abs(dpdf).max() < 1e-14 and np.abs(dcdf).max() < 1e-14:
            converged = True
        else:
            converged = False
        # update coeffs_l and coeffs_lp1 
        coeffs_lp2 = np.empty(len(coeffs_lp1) + 1)
        factor1 = (2 * l + n) / (2 * l + 4)
        factor2 = (l + n - 2) / (l + 2)
        coeffs_lp2[0] = factor1 * coeffs_lp1[1] - factor2 * coeffs_l[0]
        for k in range(1, len(coeffs_lp2) - 2):
            coeffs_lp2[k] = \
                factor1 * (coeffs_lp1[k + 1] + coeffs_lp1[k - 1]) - \
                factor2 * coeffs_l[k]
        coeffs_lp2[-1] = factor1 * coeffs_lp1[-1]
        coeffs_lp2[-2] = factor1 * coeffs_lp1[-2]
        coeffs_l = coeffs_lp1
        coeffs_lp1 = coeffs_lp2
        # update gegenbauer_l_1 and gegenbauer_lp1_1
        gegenbauer_lp1_1, gegenbauer_l_1 = \
            (2 * l + n) / (l + 2) * gegenbauer_lp1_1 - \
            (l + n - 2) / (l + 2) * gegenbauer_l_1, \
            gegenbauer_lp1_1
        # update l and the squared normalization coefficient
        l += 1
        norm_square *= 2 * l / (2 * l + 2 * n - 6)
    return pdf, cdf

@nb.jit(nopython=True, cache=True)
def angle_mean(n, var):
    """Compute the mean of the diffusion kernel on the (n-1)D sphere.

       The PDF over which to average is a function of phi_{n-1}, the 
       final polar angle in the hyperspherical coordinates of a sphere 
       embedded in nD space. The volume element, sin^{n-2}(phi_{n-1}) is 
       included, such that the PDF multiplied by the volume element 
       integrates to 1 over [0, pi].

    Parameters
    ----------
    n : int
        The dimensionality of the space in which the spherical surface is 
        embedded.
    var : np.array [N]
        Array of variance parameters for which to compute the mean of the 
        diffusion kernel.

    Returns
    -------
    mean : np.array [N]
        The mean of the diffusion kernel on the (n-1)D sphere.
    """
    norm_square = (2. / np.pi) ** (int(n - 3) % 2) * \
                  2 ** (n - 3)
    for j in range(n - 4, 0, -2):
        norm_square *= j ** 2
    for j in range(2 * n - 6, 0, -2):
        norm_square /= j
    mean = np.zeros_like(var)
    gegenbauer_l_1 = 1. # lth Gegenbauer polynomial evaluated at 1
    gegenbauer_lp1_1 = n - 2 # l+1th Gegenbauer polynomial evaluated at 1
    coeffs_l = get_cheby_coeffs(n, 0)
    coeffs_lp1 = get_cheby_coeffs(n, 1)
    converged, prev_converged = False, False
    l = 0
    while not (converged and prev_converged):
        prev_converged = converged
        dmean = np.zeros_like(var)
        for k, coeff in enumerate(coeffs_l):
            if not coeff:
                continue
            if n % 2: # n is odd; use a sine series
                dmean += coeff * (np.sin(np.pi * (k + 1)) + 
                                  np.pi * (k + 1) * np.cos(np.pi * k)) / \
                         (k + 1) ** 2
            elif k > len(coeffs_l) - n: # n is even; use a cosine series
                if k:
                    dmean += coeff * (np.pi * k * np.sin(np.pi * k) + 
                                      np.cos(np.pi * k) - 1.) / k ** 2
                else:
                    dmean += coeff * np.pi ** 2 / 2.
        exp_factor = np.exp(-0.5 * l * (l + n - 2) * var)
        dmean *= 0.5 * (2 * l + n - 2) * \
                 gegenbauer_l_1 * norm_square * exp_factor
        mean += dmean
        if np.abs(dmean).max() < 1e-14:
            converged = True
        else:
            converged = False
        # update coeffs_l and coeffs_lp1 
        coeffs_lp2 = np.empty(len(coeffs_lp1) + 1)
        factor1 = (2 * l + n) / (2 * l + 4)
        factor2 = (l + n - 2) / (l + 2)
        coeffs_lp2[0] = factor1 * coeffs_lp1[1] - factor2 * coeffs_l[0]
        for k in range(1, len(coeffs_lp2) - 2):
            coeffs_lp2[k] = \
                factor1 * (coeffs_lp1[k + 1] + coeffs_lp1[k - 1]) - \
                factor2 * coeffs_l[k]
        coeffs_lp2[-1] = factor1 * coeffs_lp1[-1]
        coeffs_lp2[-2] = factor1 * coeffs_lp1[-2]
        coeffs_l = coeffs_lp1
        coeffs_lp1 = coeffs_lp2
        # update gegenbauer_l_1 and gegenbauer_lp1_1
        gegenbauer_lp1_1, gegenbauer_l_1 = \
            (2 * l + n) / (l + 2) * gegenbauer_lp1_1 - \
            (l + n - 2) / (l + 2) * gegenbauer_l_1, \
            gegenbauer_lp1_1
        # update l and the squared normalization coefficient
        l += 1
        norm_square *= 2 * l / (2 * l + 2 * n - 6)
    return mean



import numpy as np
import numba as nb

@nb.jit(nopython=True, cache=True)
def chebyshev_first_matrix(n_max):
    M = np.zeros((n_max + 1, n_max + 1))
    for n in range(n_max + 1):
        for k in range(n + 1):
            if k % 2 != n % 2:
                continue
            M[n, k] = (-1) ** ((n - k) // 2) * factorial((n + k) // 2 - 1) / \
                      factorial((n - k) // 2) / factorial(k) * 2 ** k * n // 2
    M[0, 0] = 1
    return M

@nb.jit(nopython=True, cache=True)
def chebyshev_first_inv_matrix(n_max):
    M = np.zeros((n_max + 1, n_max + 1))
    for n in range(n_max + 1):
        for k in range(n + 1):
            if k % 2 != n % 2:
                continue
            M[n, k] = 2. ** (1 - n) * comb(n, (n - k) // 2)
            if k == 0:
                M[n, k] /= 2.
    return M

@nb.jit(nopython=True, cache=True)
def chebyshev_second_matrix(n_max):
    M = np.zeros((n_max + 1, n_max + 1))
    for n in range(n_max + 1):
        for k in range(n + 1):
            if k % 2 != n % 2:
                continue
            M[n, k] = (-1) ** ((n - k) // 2) * factorial((n + k) // 2) / \
                      factorial((n - k) // 2) / factorial(k) * 2 ** k
    return M

@nb.jit(nopython=True, cache=True)
def chebyshev_second_inv_matrix(n_max):
    M = np.zeros((n_max + 1, n_max + 1))
    for n in range(n_max + 1):
        for k in range(n + 1):
            if k % 2 != n % 2:
                continue
            M[n, k] = 2. ** -n * comb(n, (n - k) // 2) * (k + 1) / \
                      ((n + k) // 2 + 1)
    return M

@nb.jit(nopython=True, cache=True)
def factorial(n):
    fac = 1
    for i in range(1, n + 1):
        fac *= i
    return fac

@nb.jit(nopython=True, cache=True)
def comb(n, k):
    nCk = 1.
    for j in range(1, k + 1):
        nCk *= (n - k + j) / j
    return np.round(nCk)
